import asyncio

from asynctnt import Response
from tests.util import get_complex_param
from tests import BaseTarantoolTestCase


class EvalTestCase(BaseTarantoolTestCase):

    async def test__eval_basic(self):
        res = await self.conn.eval('return "hola"')

        self.assertIsInstance(res, Response, 'Got eval response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, ['hola'], 'Body ok')

    async def test__eval_basic_pack(self):
        res = await self.conn.eval('return {"hola"}')

        self.assertIsInstance(res, Response, 'Got eval response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [['hola']], 'Body ok')

    async def test__eval_with_param(self):
        args = [1, 2, 3, 'hello']
        res = await self.conn.eval('return ...', args)

        self.assertListEqual(res.body, args, 'Body ok')

    async def test__eval_with_param_pack(self):
        args = [1, 2, 3, 'hello']
        res = await self.conn.eval('return {...}', args)

        self.assertListEqual(res.body, [args], 'Body ok')

    async def test__eval_func_name_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.eval(12)

        with self.assertRaises(TypeError):
            await self.conn.eval([1, 2])

        with self.assertRaises(TypeError):
            await self.conn.eval({'a': 1})

        with self.assertRaises(TypeError):
            await self.conn.eval(b'qwer')

    async def test__eval_params_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.eval('return {...}', 220349)

        with self.assertRaises(TypeError):
            await self.conn.eval('return {...}', 'hey')

        with self.assertRaises(TypeError):
            await self.conn.eval('return {...}', {1: 1, 2: 2})

    async def test__eval_args_tuple(self):
        try:
            await self.conn.eval('return {...}', (1, 2))
        except Exception as e:
            self.fail(e)

    async def test__eval_complex_param(self):
        p, cmp = get_complex_param(encoding=self.conn.encoding)
        res = await self.conn.eval('return {...}', [p])
        self.assertDictEqual(res.body[0][0], cmp, 'Body ok')

    async def test__eval_timeout_in_time(self):
        try:
            cmd = """
            local args = {...}
            local fiber = require("fiber")
            fiber.sleep(args[1])
            """
            await self.conn.eval(cmd, [0.1], timeout=1)
        except Exception as e:
            self.fail(e)

    async def test__eval_timeout_late(self):
        cmd = """
        local args = {...}
        local fiber = require("fiber")
        fiber.sleep(args[1])
        """
        with self.assertRaises(asyncio.TimeoutError):
            await self.conn.eval(cmd, [0.3], timeout=0.1)



import asyncio

import logging

from asynctnt import Response
from asynctnt.exceptions import TarantoolDatabaseError, ErrorCode
from tests.util import get_complex_param
from tests import BaseTarantoolTestCase


class CallTestCase(BaseTarantoolTestCase):
    LOGGING_LEVEL = logging.DEBUG

    def has_new_call(self):
        return self.conn.version >= (1, 7)

    async def test__call_basic(self):
        res = await self.conn.call('func_hello')

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [['hello']], 'Body ok')

    async def test__call_basic_bare(self):
        res = await self.conn.call('func_hello_bare')
        cmp = ['hello']

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        if not self.has_new_call():
            cmp = [cmp]
        self.assertListEqual(res.body, cmp, 'Body ok')

    async def test__call_unknown_function(self):
        with self.assertRaises(TarantoolDatabaseError) as ctx:
            await self.conn.call('blablabla')
        self.assertEqual(ctx.exception.code, ErrorCode.ER_NO_SUCH_PROC)

    async def test__call_with_param(self):
        res = await self.conn.call('func_param', ['myparam'])

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertListEqual(res.body, [['myparam']], 'Body ok')

    async def test__call_with_param_bare(self):
        res = await self.conn.call('func_param_bare', ['myparam'])
        cmp = ['myparam']
        if not self.has_new_call():
            cmp = [cmp]

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertListEqual(res.body, cmp, 'Body ok')

    async def test__call_func_name_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call(12)

        with self.assertRaises(TypeError):
            await self.conn.call([1, 2])

        with self.assertRaises(TypeError):
            await self.conn.call({'a': 1})

        with self.assertRaises(TypeError):
            await self.conn.call(b'qwer')

    async def test__call_params_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call('func_param', 220349)

        with self.assertRaises(TypeError):
            await self.conn.call('func_param', 'hey')

        with self.assertRaises(TypeError):
            await self.conn.call('func_param', {1: 1, 2: 2})

    async def test__call_args_tuple(self):
        try:
            await self.conn.call('func_param', (1, 2))
        except Exception as e:
            self.fail(e)

    async def test__call_complex_param(self):
        p, cmp = get_complex_param(encoding=self.conn.encoding)
        res = await self.conn.call('func_param', [p])
        self.assertDictEqual(res.body[0][0], cmp, 'Body ok')

    async def test__call_complex_param_bare(self):
        p, cmp = get_complex_param(encoding=self.conn.encoding)
        cmp = [cmp]
        res = await self.conn.call('func_param_bare', [p])
        if not self.has_new_call():
            cmp = [cmp]
        self.assertListEqual(res.body, cmp, 'Body ok')

    async def test__call_timeout_in_time(self):
        try:
            await self.conn.call('func_long', [0.1], timeout=1)
        except Exception as e:
            self.fail(e)

    async def test__call_timeout_late(self):
        with self.assertRaises(asyncio.TimeoutError):
            await self.conn.call('func_long', [0.3], timeout=0.1)

    async def test__call_raise(self):
        with self.assertRaises(TarantoolDatabaseError) as e:
            await self.conn.call('raise')

        self.assertEqual(e.exception.code, 0, 'code by box.error{} is 0')
        self.assertEqual(e.exception.message, 'my reason', 'Reason ok')


class Call16TestCase(BaseTarantoolTestCase):
    async def test__call16_basic(self):
        res = await self.conn.call16('func_hello')

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [['hello']], 'Body ok')

    async def test__call16_basic_bare(self):
        # Tarantool automatically wraps return result into tuple

        res = await self.conn.call16('func_hello')

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [['hello']], 'Body ok')

    async def test__call16_unknown_function(self):
        with self.assertRaises(TarantoolDatabaseError) as ctx:
            await self.conn.call16('blablabla')
        self.assertEqual(ctx.exception.code, ErrorCode.ER_NO_SUCH_PROC)

    async def test__call16_with_param(self):
        res = await self.conn.call16('func_param', ['myparam'])

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertListEqual(res.body, [['myparam']], 'Body ok')

    async def test__call16_with_param_bare(self):
        # Tarantool automatically wraps return result into tuple

        res = await self.conn.call16('func_param_bare', ['myparam'])

        self.assertIsInstance(res, Response, 'Got call response')
        self.assertListEqual(res.body, [['myparam']], 'Body ok')

    async def test__call16_func_name_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call16(12)

        with self.assertRaises(TypeError):
            await self.conn.call16([1, 2])

        with self.assertRaises(TypeError):
            await self.conn.call16({'a': 1})

        with self.assertRaises(TypeError):
            await self.conn.call16(b'qwer')

    async def test__call16_params_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call16('func_param', 220349)

        with self.assertRaises(TypeError):
            await self.conn.call16('func_param', 'hey')

        with self.assertRaises(TypeError):
            await self.conn.call16('func_param', {1: 1, 2: 2})

    async def test__call16_args_tuple(self):
        try:
            await self.conn.call16('func_param', (1, 2))
        except Exception as e:
            self.fail(e)

    async def test__call16_complex_param(self):
        p, cmp = get_complex_param(encoding=self.conn.encoding)
        res = await self.conn.call('func_param', [p])
        self.assertDictEqual(res.body[0][0], cmp, 'Body ok')

    async def test__call16_complex_param_bare(self):
        p, cmp = get_complex_param(encoding=self.conn.encoding)
        res = await self.conn.call16('func_param_bare', [p])
        self.assertDictEqual(res.body[0][0], cmp, 'Body ok')

    async def test__call16_timeout_in_time(self):
        try:
            await self.conn.call16('func_long', [0.1], timeout=1)
        except Exception as e:
            self.fail(e)

    async def test__call_timeout_late(self):
        with self.assertRaises(asyncio.TimeoutError):
            await self.conn.call16('func_long', [0.3], timeout=0.1)



from libc.stdint cimport int64_t, uint32_t


cdef class TntField:
    cdef:
        uint32_t id
        str name
        str type

    @staticmethod
    cdef TntField new(uint32_t id, str name, str type)


cdef class SchemaIndex:
    cdef:
        int sid
        int iid
        str name
        str index_type
        object unique
        list parts

        list fields

    @staticmethod
    cdef SchemaIndex new()


cdef class SchemaDummyIndex(SchemaIndex):
    pass


cdef class SchemaSpace:
    cdef:
        int sid
        int owner
        str name
        str engine
        int field_count
        object flags

        list fields
        dict fields_map
        dict indexes

    @staticmethod
    cdef SchemaSpace new()

    cdef add_index(self, SchemaIndex idx)
    cdef SchemaIndex get_index(self, index, create_dummy=*)


cdef class SchemaDummySpace(SchemaSpace):
    pass


cdef class Schema:
    cdef:
        dict schema
        int64_t id

    @staticmethod
    cdef Schema new(int64_t schema_id)

    cdef SchemaSpace get_space(self, space)
    cdef SchemaSpace create_dummy_space(self, int space_id)
    cdef SchemaSpace get_or_create_space(self, space)

    cdef SchemaSpace parse_space(self, list index_row)
    cdef SchemaIndex parse_index(self, list index_row)

    cdef inline clear(self)

    @staticmethod
    cdef Schema parse(int64_t schema_id, spaces, indexes)


cdef list dict_to_list_fields(list fields, dict d, bint default_none)



from libc.stdint cimport uint32_t, uint64_t, int64_t

cimport tnt


cdef inline size_t size_t_max(size_t a, size_t b)
cdef inline uint32_t nearest_power_of_2(uint32_t v)


cdef class ReadBuffer:
    cdef:
        char *buf
        size_t initial_buffer_size  # Initial buffer size, obviously
        size_t len  # Allocated size
        size_t use  # Used size

        str encoding

    @staticmethod
    cdef ReadBuffer new(str encoding, size_t initial_buffer_size=*)

    cdef void _reallocate(self, size_t new_size) except *
    cdef int extend(self, const char *data, size_t len) except -1
    cdef void move(self, size_t pos)
    cdef void move_offset(self, ssize_t offset, size_t size) except *
    cdef bytes get_slice(self, size_t begin, size_t end)
    cdef bytes get_slice_begin(self, size_t begin)
    cdef bytes get_slice_end(self, size_t end)



from libc.stdint cimport uint64_t, uint32_t, int64_t


cdef class Response:
    cdef:
        uint32_t _code
        uint32_t _return_code
        uint64_t _sync
        int64_t _schema_id
        str _errmsg
        list _body
        bytes _encoding
        Request _req

    @staticmethod
    cdef inline Response new(bytes encoding)

    cdef inline is_error(self)


cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,
                                   Response resp) except -1
cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,
                                 Response resp) except -1



cdef enum tnt_header_key:
    TP_CODE = 0x00
    TP_SYNC = 0x01
    TP_SERVER_ID = 0x02
    TP_LSN = 0x03
    TP_TIMESTAMP = 0x04
    TP_SCHEMA_ID = 0x05


cdef enum tnt_body_key_t:
    TP_SPACE = 0x10
    TP_INDEX = 0x11
    TP_LIMIT = 0x12
    TP_OFFSET = 0x13
    TP_ITERATOR = 0x14
    TP_KEY = 0x20
    TP_TUPLE = 0x21
    TP_FUNCTION = 0x22
    TP_USERNAME = 0x23
    TP_EXPRESSION = 0x27
    TP_OPERATIONS = 0x28


cdef enum tnt_response_key_t:
    TP_DATA = 0x30
    TP_ERROR = 0x31


cdef enum tp_request_type:
    TP_SELECT = 0x01
    TP_INSERT = 0x02
    TP_REPLACE = 0x03
    TP_UPDATE = 0x04
    TP_DELETE = 0x05
    TP_CALL_16 = 0x06
    TP_CALL = 0x0a
    TP_AUTH = 0x07
    TP_EVAL = 0x08
    TP_UPSERT = 0x09
    TP_PING = 0x40


cdef enum tnt_update_op_kind:
    OP_UPD_ARITHMETIC
    OP_UPD_DELETE
    OP_UPD_INSERT_ASSIGN
    OP_UPD_SPLICE
    OP_UPD_UNKNOWN



from asynctnt.exceptions import TarantoolSchemaError
from asynctnt.log import logger
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.stdint cimport int32_t
cimport cpython
cimport cpython.dict
cimport cpython.list
from cpython.ref cimport PyObject

cdef class TntField:
    @staticmethod
    cdef TntField new(uint32_t id, str name, str type):
        cdef TntField f
        f = TntField.__new__(TntField)
        f.id = id
        f.name = name
        f.type = type
        return f


cdef class SchemaIndex:
    @staticmethod
    cdef SchemaIndex new():
        cdef SchemaIndex idx
        idx = SchemaIndex.__new__(SchemaIndex)
        idx.iid = -1
        idx.sid = -1
        idx.name = None
        idx.index_type = None
        idx.unique = None
        idx.parts = []
        idx.fields = []
        return idx

    def __repr__(self):
        return \
            '<{} sid={}, id={}, name={}, ' \
            'type={}, unique={}>'.format(
                self.__class__.__name__,
                self.sid, self.iid, self.name, self.index_type, self.unique
            )


cdef class SchemaDummyIndex(SchemaIndex):
    pass


cdef class SchemaSpace:

    @staticmethod
    cdef SchemaSpace new():
        cdef SchemaSpace sp
        sp = SchemaSpace.__new__(SchemaSpace)
        sp.sid = -1
        sp.owner = -1
        sp.name = None
        sp.engine = None
        sp.field_count = 0
        sp.flags = None

        sp.fields = []
        sp.fields_map = {}
        sp.indexes = {}
        return sp

    cdef add_index(self, SchemaIndex idx):
        cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)
        if idx.name:
            cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)

    cdef SchemaIndex get_index(self, index, create_dummy=True):
        cdef:
            SchemaIndex idx
            bint is_str, is_int
            PyObject *obj_p
        is_str = isinstance(index, str)
        is_int = isinstance(index, int)
        if not is_str and not is_int:
            raise TypeError(
                'Index must be either str or int, got: {}'.format(type(index)))

        obj_p = cpython.dict.PyDict_GetItem(self.indexes, index)
        if obj_p is not NULL:
            return <SchemaIndex>obj_p
        else:
            if is_int and create_dummy:
                logger.debug(
                    'Index %s not found in space %s/%s. Creating dummy.',
                    index, self.sid, self.name
                )
                idx = SchemaDummyIndex.new()
                idx.iid = index
                idx.sid = self.sid
                idx.name = str(index)
                cpython.dict.PyDict_SetItem(self.indexes, index, idx)
                return idx
            else:
                raise TarantoolSchemaError(
                    'Index {} not found in space {}/{}'.format(
                        index, self.sid, self.name
                    )
                )

    def __repr__(self):
        return '<{} id={}, name={}, arity={}>'.format(
            self.__class__.__name__,
            self.sid, self.name, self.arity
        )


cdef class SchemaDummySpace(SchemaSpace):
    pass


cdef class Schema:
    @staticmethod
    cdef Schema new(int64_t schema_id):
        cdef Schema s
        s = Schema.__new__(Schema)
        s.id = schema_id
        s.schema = {}
        return s

    cdef SchemaSpace get_space(self, space):
        cdef PyObject *obj_p = \
            cpython.dict.PyDict_GetItem(self.schema, space)
        if obj_p is NULL:
            return None
        return <SchemaSpace>obj_p

    cdef SchemaSpace get_or_create_space(self, space):
        cdef:
            bint is_str, is_int
            PyObject *obj_p
        is_str = isinstance(space, str)
        is_int = isinstance(space, int)
        if not is_str and not is_int:
            raise TypeError(
                'Space must be either str or int, got: {}'.format(type(space)))

        obj_p = cpython.dict.PyDict_GetItem(self.schema, space)
        if obj_p is NULL:
            if is_str:
                raise TarantoolSchemaError(
                    'Space {} not found'.format(space)
                )
            else:
                return self.create_dummy_space(space)
        return <SchemaSpace>obj_p

    cdef SchemaSpace create_dummy_space(self, int space_id):
        cdef SchemaSpace s
        logger.debug('Space %s not found. Creating dummy.', space_id)
        s = SchemaDummySpace.new()
        s.sid = space_id
        s.name = str(space_id)
        cpython.dict.PyDict_SetItem(self.schema, space_id, s)
        return s

    cdef inline clear(self):
        self.schema.clear()

    cdef SchemaSpace parse_space(self, list space_row):
        cdef:
            SchemaSpace sp
            size_t k
            size_t row_len
            list format_list
            TntField f

        assert space_row is not None

        sp = SchemaSpace.new()
        row_len = <size_t>cpython.list.PyList_GET_SIZE(space_row)

        k = 0
        sp.sid = <int><object>cpython.list.PyList_GET_ITEM(space_row, k)
        k += 1
        if k < row_len:
            sp.owner = <int><object>cpython.list.PyList_GET_ITEM(space_row, k)
        k += 1
        if k < row_len:
            sp.name = <str>cpython.list.PyList_GET_ITEM(space_row, k)
        k += 1
        if k < row_len:
            sp.engine = <str>cpython.list.PyList_GET_ITEM(space_row, k)
        k += 1
        if k < row_len:
            sp.field_count = <int><object>cpython.list.PyList_GET_ITEM(
                space_row, k)
        k += 1
        if k < row_len:
            sp.flags = <object>cpython.list.PyList_GET_ITEM(space_row, k)
        k += 1
        if k < row_len:
            # format
            format_list = <list>cpython.list.PyList_GET_ITEM(space_row, k)
            for i in range(len(format_list)):
                field_id = i
                field_name = format_list[i]['name']
                field_type = format_list[i]['type']
                f = TntField.new(field_id, field_name, field_type)

                cpython.list.PyList_Append(sp.fields, field_name)
                cpython.dict.PyDict_SetItem(sp.fields_map, field_name, f)

        return sp

    cdef SchemaIndex parse_index(self, list index_row):
        cdef:
            SchemaIndex idx
            SchemaSpace sp
            uint32_t i

        assert index_row is not None

        idx = SchemaIndex.new()
        idx.sid = <int><object>cpython.list.PyList_GetItem(index_row, 0)
        idx.iid = <int><object>cpython.list.PyList_GetItem(index_row, 1)
        idx.name = <str>cpython.list.PyList_GetItem(index_row, 2)
        idx.index_type = <str>cpython.list.PyList_GetItem(index_row, 3)
        idx.unique = <object>cpython.list.PyList_GetItem(index_row, 4)
        idx.parts = []
        idx.fields = []

        sp = self.get_space(idx.sid)
        if sp is None:
            raise TarantoolSchemaError(
                'Space with id {} not found'.format(idx.sid))

        parts = index_row[5]
        if isinstance(parts, (list, tuple)):
            for field_id, field_type in parts:
                cpython.list.PyList_Append(
                    idx.parts,
                    (field_id, field_type)
                )

                if field_id < len(sp.fields):
                    cpython.list.PyList_Append(
                        idx.fields,
                        <str>cpython.list.PyList_GetItem(sp.fields, field_id)
                    )
                else:
                    logger.warning(
                        'Field #%d of space %s is not '
                        'in space format definition', field_id, sp.name)
        else:
            for i in range(index_row[5]):
                field_id = index_row[5 + 1 + i * 2]
                field_type = index_row[5 + 2 + i * 2]

                cpython.list.PyList_Append(
                    idx.parts,
                    (field_id, field_type)
                )

                if field_id < len(sp.fields):
                    cpython.list.PyList_Append(
                        idx.fields,
                        <str>cpython.list.PyList_GetItem(sp.fields, field_id)
                    )
                else:
                    logger.warning(
                        'Field #%d of space %s is not '
                        'in space format definition', field_id, sp.name)

        return idx

    @staticmethod
    cdef Schema parse(int64_t schema_id, spaces, indexes):
        cdef:
            Schema s
            SchemaSpace sp
            SchemaIndex idx

        s = Schema.new(schema_id)
        for space_row in spaces:
            sp = s.parse_space(space_row)
            cpython.dict.PyDict_SetItem(s.schema, sp.sid, sp)
            if sp.name:
                cpython.dict.PyDict_SetItem(s.schema, sp.name, sp)

        for index_row in indexes:
            idx = s.parse_index(index_row)
            sp = s.schema[idx.sid]
            sp.add_index(idx)
        return s

    def __repr__(self):
        return '<Schema>'


cdef list dict_to_list_fields(list fields, dict d, bint default_none):
    cdef:
        list l
        dict used
        object value, field
        int32_t used_count
        uint32_t fields_count, field_id
        PyObject *obj_p

    assert d is not None

    l = []
    used_count = <int32_t>cpython.dict.PyDict_Size(d)
    fields_count = <uint32_t>cpython.list.PyList_GET_SIZE(fields)

    for field_id in range(fields_count):
        field = <object>cpython.list.PyList_GetItem(fields, field_id)

        obj_p = cpython.dict.PyDict_GetItem(d, field)
        if obj_p is not NULL:
            value = <object>obj_p
            used_count -= 1
            l.append(value)
        else:
            if default_none:
                l.append(None)

    # Warn user if he used any of unknown fields
    if used_count != 0:
        used = {}
        for field_id in range(fields_count):
            field = <object>cpython.list.PyList_GetItem(fields, field_id)

            if <bint>cpython.dict.PyDict_Contains(d, field):
                used[field] = None
        if <bint>cpython.dict.PyDict_Contains(d, ''):
            used[''] = None

        for field in d:
            if field not in used:
                logger.warning(
                    'Field \'%s\' in supplied dict is unknown as '
                    'a tuple field for selected index. Skipping.',
                    field
                )
    return l



from libc.stdint cimport uint64_t, uint32_t, int64_t


cdef class Db:
    cdef:
        BaseProtocol _protocol
        bytes _encoding

    @staticmethod
    cdef inline Db new(BaseProtocol protocol)

    cdef inline uint64_t next_sync(self)
    # cdef execute(self, Request req, float timeout)

    cdef Request _ping(self)
    cdef Request _call16(self, str func_name, args)
    cdef Request _call(self, str func_name, args)
    cdef Request _eval(self, str expression, args)
    cdef Request _select(self, SchemaSpace space, SchemaIndex index, key,
                         uint64_t offset, uint64_t limit, uint32_t iterator)
    cdef Request _insert(self, SchemaSpace space, t, bint replace)
    cdef Request _delete(self, SchemaSpace space, SchemaIndex index, key)
    cdef Request _update(self, SchemaSpace space, SchemaIndex index, key,
                         list operations)
    cdef Request _upsert(self, SchemaSpace space, t, list operations)
    cdef Request _auth(self, bytes salt, str username, str password)

    @staticmethod
    cdef bytes _sha1(tuple values)

    @staticmethod
    cdef bytes _strxor(bytes hash1, bytes scramble)



cimport cpython.unicode
cimport cpython.list
cimport cpython.dict

from libc.stdint cimport uint32_t, uint64_t, int64_t

cimport tnt

from asynctnt.log import logger
import yaml


cdef class Response:
    """
        Response object for all the requests to Tarantool
    """
    def __cinit__(self):
        self._sync = 0
        self._code = 0
        self._return_code = 0
        self._schema_id = -1
        self._errmsg = None
        self._body = None
        self._encoding = None
        self._req = None

    @staticmethod
    cdef inline Response new(bytes encoding):
        cdef Response resp
        resp = Response.__new__(Response)
        resp._encoding = encoding
        return resp

    cdef inline is_error(self):
        return self._code != 0

    def __repr__(self):
        body_len = None
        if self._body is not None:
            body_len = len(self._body)
        return '<Response: code={}, sync={}, body_len={}>'.format(
            self._code, self._sync, body_len)

    def body2yaml(self):
        """
            Dumps body to YAML (for pretty formatting)
            :return: YAML-formatted string
        """
        return yaml.dump(self._body, allow_unicode=True)

    @property
    def sync(self):
        """
            Response's sync (incremental id) for the corresponding request
        """
        return self._sync

    @property
    def code(self):
        """
            Response code (0 - success)
        """
        return self._code

    @property
    def return_code(self):
        """
            Response return code (It's essentially a code & 0x7FFF)
        """
        return self._return_code

    @property
    def schema_id(self):
        """
            Current scema id in Tarantool
        """
        return self._schema_id

    @property
    def errmsg(self):
        """
            If self.code != 0 then errmsg contains an error message
        """
        return self._errmsg

    @property
    def body(self):
        """
            Return body. Always a list.
        """
        return self._body

    @property
    def encoding(self):
        """
            Response encoding
        """
        return self._encoding


cdef object _decode_obj(const char **p, bytes encoding):
    cdef:
        uint32_t i
        mp_type obj_type

        const char *s
        uint32_t s_len

        uint32_t arr_size
        list arr

        uint32_t map_size
        dict map
        mp_type map_key_type
        const char *map_key_str
        uint32_t map_key_len
        object map_key

    obj_type = mp_typeof(p[0][0])
    if obj_type == MP_UINT:
        return mp_decode_uint(p)
    elif obj_type == MP_INT:
        return mp_decode_int(p)
    elif obj_type == MP_STR:
        s = NULL
        s_len = 0
        s = mp_decode_str(p, &s_len)
        return decode_string(s[:s_len], encoding)
    elif obj_type == MP_BIN:
        s = NULL
        s_len = 0
        s = mp_decode_bin(p, &s_len)
        return <bytes>s[:s_len]
    elif obj_type == MP_BOOL:
        return mp_decode_bool(p)
    elif obj_type == MP_FLOAT:
        return mp_decode_float(p)
    elif obj_type == MP_DOUBLE:
        return mp_decode_double(p)
    elif obj_type == MP_ARRAY:
        arr_size = mp_decode_array(p)
        value = []
        for i in range(arr_size):
            value.append(_decode_obj(p, encoding))
        return value
    elif obj_type == MP_MAP:
        map = {}
        map_size = mp_decode_map(p)
        for i in range(map_size):
            map_key_type = mp_typeof(p[0][0])
            if map_key_type == MP_STR:
                map_key_len = 0
                map_key_str = mp_decode_str(p, &map_key_len)
                map_key = <object>(
                    decode_string(map_key_str[:map_key_len], encoding)
                )
            elif map_key_type == MP_UINT:
                map_key = <object>mp_decode_uint(p)
            elif map_key_type == MP_INT:
                map_key = <object>mp_decode_int(p)
            else:
                mp_next(p)  # skip current key
                mp_next(p)  # skip value
                logger.warning('Unexpected key type in map: %s',
                               map_key_type)
                continue

            map[map_key] = _decode_obj(p, encoding)

        return map
    elif obj_type == MP_NIL:
        mp_next(p)
        return None
    else:
        mp_next(p)
        logger.warning('Unexpected obj type: %s', obj_type)
        return None


cdef list _response_parse_body_data(const char **b, Response resp):
    cdef:
        uint32_t size
        uint32_t tuple_size, max_tuple_size
        list tuples
        uint32_t i, k

        Request req
        dict d
        str field_name
        object value
        list fields

    size = mp_decode_array(b)
    tuples = []
    req = resp._req
    if req is not None \
            and req.tuple_as_dict \
            and req.space is not None \
            and req.space.fields is not None:
        # decode as dicts

        fields = req.space.fields
        max_tuple_size = <uint32_t>cpython.list.PyList_GET_SIZE(fields)

        for i in range(size):
            if mp_typeof(b[0][0]) != MP_ARRAY:
                raise TypeError(
                    'Tuple must be an array when decoding as dict'
                )
            d = {}
            unknown_fields = None
            tuple_size = mp_decode_array(b)
            for k in range(tuple_size):
                value = _decode_obj(b, resp._encoding)
                if k < max_tuple_size:
                    field_name = <str>cpython.list.PyList_GET_ITEM(fields, k)
                    d[field_name] = value
                else:
                    if unknown_fields is None:
                        unknown_fields = []
                    unknown_fields.append(value)

            if unknown_fields is not None:
                d[''] = unknown_fields
            tuples.append(d)
    else:
        for i in range(size):
            tuples.append(_decode_obj(b, resp._encoding))

    return tuples


cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,
                                   Response resp) except -1:
    cdef:
        const char *b
        uint32_t size
        uint32_t key

    b = <const char*>buf
    if mp_typeof(b[0]) != MP_MAP:
        raise TypeError('Response header must be a MP_MAP')

    # parsing header
    size = mp_decode_map(&b)
    for _ in range(size):
        if mp_typeof(b[0]) != MP_UINT:
            raise TypeError('Header key must be a MP_UINT')

        key = mp_decode_uint(&b)
        if key == tnt.TP_CODE:
            if mp_typeof(b[0]) != MP_UINT:
                raise TypeError('code type must be a MP_UINT')

            resp._code = <uint32_t>mp_decode_uint(&b)
            resp._return_code = resp._code & 0x7FFF
        elif key == tnt.TP_SYNC:
            if mp_typeof(b[0]) != MP_UINT:
                raise TypeError('sync type must be a MP_UINT')

            resp._sync = mp_decode_uint(&b)
        elif key == tnt.TP_SCHEMA_ID:
            if mp_typeof(b[0]) != MP_UINT:
                raise TypeError('schema_id type must be a MP_UINT')

            resp._schema_id = mp_decode_uint(&b)
        else:
            logger.warning('Unknown argument in header. Skipping.')
            mp_next(&b)

    return <ssize_t>(b - buf)


cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,
                                 Response resp) except -1:
    cdef:
        const char *b
        uint32_t size
        uint32_t key
        uint32_t s_len
        const char *s

    b = <const char*>buf

    # parsing body
    if b == &buf[buf_len]:
        # buffer exceeded
        return 0

    if mp_typeof(b[0]) != MP_MAP:
        raise TypeError('Response body must be a MP_MAP')

    size = mp_decode_map(&b)
    for _ in range(size):
        if mp_typeof(b[0]) != MP_UINT:
            raise TypeError('Header key must be a MP_UINT')

        key = mp_decode_uint(&b)
        if key == tnt.TP_ERROR:
            if mp_typeof(b[0]) != MP_STR:
                raise TypeError('errstr type must be a MP_STR')

            s = NULL
            s_len = 0
            s = mp_decode_str(&b, &s_len)
            resp._errmsg = decode_string(s[:s_len], resp._encoding)
        elif key == tnt.TP_DATA:
            if mp_typeof(b[0]) != MP_ARRAY:
                raise TypeError('body data type must be a MP_ARRAY')
            resp._body = _response_parse_body_data(&b, resp)

    return <ssize_t>(b - buf)



cimport cython
from libc.stdint cimport uint64_t, int64_t

cimport tnt


@cython.final
@cython.freelist(_BUFFER_FREELIST_SIZE)
cdef class Request:
    @staticmethod
    cdef inline Request new(tnt.tp_request_type op,
                            uint64_t sync, int64_t schema_id,
                            WriteBuffer buf, SchemaSpace space):
        cdef Request req
        req = Request.__new__(Request)
        req.op = op
        req.sync = sync
        req.schema_id = schema_id
        req.buf = buf
        req.tuple_as_dict = False
        req.space = space
        req.waiter = None
        req.timeout_handle = None
        return req

    def __repr__(self):
        return '<Request op={}, sync={}, schema_id={}, len={}>'.format(
            self.op,
            self.sync,
            self.schema_id,
            self.buf.len()
        )
